/*
Very simple GLUT program using bare OpenGL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2006/10/05 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#include "ogl/readback.h"
#include "ogl/readback.cpp"

#include "osl/random.h"
#include "osl/random_vector.h"
#include "osl/random.cpp" /* For easy linking */

#ifndef M_PI
# define M_PI           3.14159265358979323846  /* pi */
#endif


/* Variables updated by the GUI */
float cameraZoom=0.5, cameraDist=3.0;
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */
int view_w, view_h; /**< Width and height (in pixels) of our viewport at last resize */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 gui_modelMat(1.0), gui_viewMat(1.0), light1Mat(1.0), light2Mat(1.0);

/* The thing to draw (sequence of glVertex calls) */
int model_mode=0; /* updated by listbox, below */
enum {MODEL_SPHERE, MODEL_TEAPOT, MODEL_TORUS, MODEL_PLANE};
const char *model_mode_names[]={"Sphere", "Teapot", "Torus", "Plane",0};
int model_anim=0; /* updated by listbox, below */
enum {ANIMATE_OFF,ANIMATE_FAST,ANIMATE_SLOW};
const char *model_anim_names[]={"Off", "Fast", "Slow",0};


/* Our shadow method */
int shadow_mode=4; /* updated by listbox, below */
enum {SHADOW_NONE,SHADOW_DECAL, 
	SHADOW_PROJECTED, SHADOW_PROJECTED_STENCIL, 
	SHADOW_VOLUMES, SHADOW_MAPS, SHADOW_PENUMBRA,
};
const char *shadow_mode_names[]={"None","Decal", 
	"Projected (no stencil)", "Projected (stencil)", 
	"Shadow Volumes", "Shadow Maps", "Penumbra Limit Maps",
	0
};

int shadowmap_res=0;
const char *shadowmap_res_names[]={"High","Med","Low",
	0
};


/* Shows off various details of the shadow method */
int shadow_detail1=0;
int shadow_detail2=0;
int shadow_detail3=0;

GLUI_Checkbox *check_detail1, *check_detail2, *check_detail3;

void setname(GLUI_Checkbox *c,const char *n) {
	if (c->name!=n) c->set_name(n);
}
void set_shadow_detail_names(const char *n1,const char *n2="",const char *n3="") {
	setname(check_detail1,n1);
	setname(check_detail2,n2);
	setname(check_detail3,n3);
}


void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATALITY: OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************* Objects we'll be drawing ***********/
class drawObject; /* forward declaration */
class shadowMethod; /* forward declaration */

// Information about everything that can be drawn
class drawWorld {
public:
	vec3 lightDir; /* (unit) direction light is arriving on the scene */
	mat4 lightFlatten; /* Matrix to flatten world coordinates onto ground plane, moving along light rays */
	vec3 lightDir2;
	mat4 lightFlatten2;
	
	double groundZ; /* Height of the ground, on the Z axis. */
	
	/**
	 Coordinate systems used here:
	
	Model coordinates, which are those passed in to glVertex
		modelMat transforms to world coordinates
	World coordinates, where lighting happens
		viewMat transforms to eye coordinates
	Eye coordinates, where the camera lies on the Z axis
		persMat transforms to clip coordinates
	Clip coordinates, onscreen position from -1 to +1 in X and Y
		the graphics hardware internally transforms clip coordinates to
	Window coordinates, which are just onscreen pixels.
	
	The vertex shader reads model coordinates in gl_Vertex
	and writes clip coordinates in gl_Position.
	*/
	mat4 modelMat; /* model matrix--world from model coordinates */
	mat4 viewMat; /* view matrix--camera from world coordinates */
	mat4 persMat; /* perspective matrix--clip from camera coordinates (and divide by W) */
	
	// "Readback" object, used to copy the screen into a texture
	oglReadback *rb;
	
	// Objects to draw
	std::vector<drawObject> objects;
	
	void initObjects(void);
	
	void drawScene(void) const;
	void drawObjects(void) const;
	void drawGround(void) const;

	drawWorld();
	void draw(shadowMethod &shadow);
};


// A way of rendering the world with shadows
class shadowMethod {
public:
	virtual ~shadowMethod();
	// Draw this world with your shadows 
	virtual void draw(const drawWorld &w) =0;
};

// Receives pieces of a shadow silhouette
class silhoutteConsumer {
public:
	virtual ~silhoutteConsumer();
	// Here's the first point on a shadow silhoutte.
	//   world is the vertex location in world coords
	virtual void silhoutte_first(const vec3 &world) =0;
	// Here's the next point in a shadow silhoutte.
	virtual void silhoutte_next(const vec3 &world) =0;
};


// Information about a single particle
class drawObject {
public:
	double radius;
	osl::Vector3d center;
	osl::Vector3d velocity;
	osl::Vector3d color;
	osl::Axes3d orientation;
	osl::Vector3d angVel;
	drawObject(osl::Random &r)
		:radius(0.1+r.nextFloat()*0.1),
		center(osl::nextCubeVector3d(r)+osl::Vector3d(0,0,1+radius)),
		velocity(osl::nextDirection(r)),
		color(osl::nextCubeVector3d(r)),
		angVel(osl::nextCubeVector3d(r))
	{
		orientation.nudge((r.nextFloat()-0.5)*2,(r.nextFloat()-0.5)*2);
		orientation.rotate((r.nextFloat()-0.5)*2);
	}
	void draw(const drawWorld &w) const;
	// Extract this object's silhoutte points--the places where the surface is perpendicular to perpDir.
	void getSilhouette(const drawWorld &w,const vec3 &perpDir,silhoutteConsumer &dest) const;
	void animate(const drawWorld &w,double dt);
};

/*********************** DrawWorld *****************/
drawWorld::drawWorld() 
{
	groundZ=0.0; /* Z-height of ground plane */
	rb=new oglReadback;
	initObjects();
}

void drawWorld::initObjects(void) {
	objects.erase(objects.begin(),objects.end());
	osl::Random48 r;
	for (unsigned int i=0;i<20;i++) 
		objects.push_back(drawObject(r));
}


/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. 
*/
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}

/** Shift the output coordinates of this matrix by this vector */
void translate(mat4 &m,const vec4 &v) {
	for (int i=0;i<4;i++)
		m[3][i]+=v[i];
}

/** Return a "z-flatten" matrix to squish all geometry onto this Z plane,
along this direction.  Rationale: this is the matrix form of:
	double del_z=to_z-v.z;
	v.x+=del_z*lightDir.x/lightDir.z;
	v.y+=del_z*lightDir.y/lightDir.z;
	v.z+=del_z; 
*/
mat4 make_flatten(double to_z,vec3 along_dir)
{
	/* Scale so along_dir.z==1.0 */
	along_dir=along_dir/along_dir.z;
	
	mat4 flat=mat4(1.0);
	flat[2] += vec4(-along_dir,0.0); /* scale incoming Z by -along_dir */
	flat[3] += vec4(to_z*along_dir,0.0); /* and translate so z==to_z */
	return flat;
}

/** Main drawing routine, called by main_display below */
void drawWorld::draw(shadowMethod &shadow)
{
	oglCheck("start draw");
	double new_z=groundZ+0.002;

/* Set up lights */
	lightDir=vec3(light1Mat[2]); /* Z axis of light matrix is light direction */
	lightFlatten=make_flatten(new_z,lightDir);
	lightDir2=vec3(light2Mat[2]); /* Z axis of light matrix is light direction */
	lightFlatten2=make_flatten(new_z,lightDir2);
	
/* Set up transform matrices */
	modelMat=gui_modelMat; /* read rotation part of model matrix */
	
	// Figure out where the camera is supposed to be
	viewMat=gui_viewMat;
	/* Center camera on spheres */
	translate(viewMat,-viewMat*vec4(0.5,0.5,0.7,0.0));
	
	vec4 cameraLoc=vec4(0,0,cameraDist,1.0); // Location of camera after viewMat, and before perspective matrix
	mat4 viewInv=transpose(viewMat); // inverse is transpose for rotation matrix
	cameraLoc=viewInv*cameraLoc; // Location of camera in world coordinates
	//p->set("cameraLoc",cameraLoc);
	
	// Perspective matrix
	double z=cameraZoom*cameraDist;
	persMat=mat4(1.0);
	persMat[0].x=z*view_h/(float)view_w; /* aspect ratio correction */
	persMat[1].y=z;
	persMat[2].z=-0.1; /* FIXME: set depth buffer range in some smarter way... */
	persMat[2].w=-1.0; /* Flip Z axis, so -Z is deeper into screen (bigger W) */
	persMat[3].w=cameraDist;
	
	// Overall coordinate transform: perspective  view   model
	mat4 clipMat=persMat*viewMat;
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(&clipMat[0][0]);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
/* Set up pixel shader, and draw the main geometry */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	static oglProgramObject *simpleP=oglProgramFromStrings(
		"/* Vertex program */\n"
		"uniform mat4 clipFmWorld;\n"
		"varying vec4 color;\n"
		"varying vec3 normal;\n"
		"void main(void) { \n"
		"	vec4 world=gl_ModelViewMatrix*gl_Vertex; /* world coords */ \n"
		"	gl_Position=clipFmWorld*world; \n"
		"	color=gl_Color;\n"
		"	normal=normalize(gl_NormalMatrix*gl_Normal);\n"
		"}\n"
	,
		"/* Fragment program-- usual per-pixel lighting */\n"
		"varying vec4 color;\n"
		"varying vec3 normal;\n"
		"uniform vec4 lightDir, lightDir2;\n"
		"void main(void) { \n"
		"	vec3 N=normalize(normal);\n"
		"	float dotNL=clamp(dot(N,vec3(lightDir)),0.0,1.0);\n"
		"	float dotNL2=clamp(dot(N,vec3(lightDir2)),0.0,1.0);\n"
		"	gl_FragColor=color*(0.1+0.4*dotNL+0.4*dotNL2);\n"
		"	gl_FragColor.a=1.0;\n"
		"}\n"
	);
	simpleP->begin();
	simpleP->set("clipFmWorld",persMat*viewMat); /* normal camera viewpoint */
	simpleP->set("lightDir",lightDir);
	simpleP->set("lightDir2",lightDir2);
	oglCheck("after uploading simple program");

	shadow.draw(*this);
	
	simpleP->end();
	

/* Animate objects */
	static double lastAnim=oglTimer();
	double del=oglTimer()-lastAnim; /* animation step size is proportional to frame time */
	if (del>0.1) del=0.1;
	lastAnim=oglTimer();
	del*=1.0; /* set overall time scale */
	switch (model_anim) {
	  case ANIMATE_OFF: del=0; break; /* do nothing */
	  case ANIMATE_FAST: del*=1.0; break; /* Normal time */
	  case ANIMATE_SLOW: del*=0.01; break; /* Bullet time */
	};
	if (del>0) {
		for (unsigned int i=0;i<objects.size();i++) {
			objects[i].animate(*this,del);
		}
	}
	if (oglToggles['R']) { /* reset objects */
		initObjects();
		oglToggles['R']=0;
	}
	oglRepaint(10);
}
/* Draw all the geometry in the scene */
void drawWorld::drawScene(void) const
{
	drawGround();
	drawObjects();
}

void drawWorld::drawGround(void) const {
	glColor3f(0.6,0.6,0.6);
	glNormal3f(0,0,1);
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUAD_STRIP);
	double s=20.0; // ground size
	glVertex3f(-s,-s,0);
	glVertex3f(1+s,-s,0);
	glVertex3f(-s,1+s,0);
	glVertex3f(1+s,1+s,0);
	glEnd();
}
void drawWorld::drawObjects(void) const {
	/* Rotate all objects by the model matrix */
	glLoadMatrixf(&modelMat[0][0]);
	for (unsigned int i=0;i<objects.size();i++) 
		objects[i].draw(*this);
	/* Restore the model matrix */
	glLoadIdentity();
}

/********************** drawObject *****************/

/* Sphere drawn as a set of points */
void drawSphere(void)
{ 
	glutSolidSphere(1.0,10,5);
}

/* Draw a torus */
void drawTorus(void) {
	glutSolidTorus(0.2,0.5,40,60);
}
void drawTeapot(void) {
	glutSolidTeapot(1.0);
}

/* Draw a plane */
void drawPlane(void) {
	glBegin(GL_TRIANGLE_FAN);
	glNormal3f(0,0,1);
	glVertex2f(-1,-1);
	glVertex2f(+1,-1);
	glVertex2f(+1,+1);
	glVertex2f(-1,+1);
	glEnd();
}

void drawObject::draw(const drawWorld &w) const
{
	glPushMatrix();
	glTranslated(center.x,center.y,center.z);
	
	/* Copy the particle orientation into the current matrix */
	double m[16]={0.0};
	for (int axis=0;axis<3;axis++) {
		m[4*axis+0]=orientation.getX()[axis]; 
		m[4*axis+1]=orientation.getY()[axis]; 
		m[4*axis+2]=orientation.getZ()[axis]; 
		m[4*axis+3]=0.0;
	}
	m[15]=1.0;
	glMultMatrixd(m);
	
	glScalef(radius,radius,radius);
	glDisable(GL_TEXTURE_2D);
	glColor3f(color.x,color.y,color.z);
	switch(model_mode) {
	case MODEL_SPHERE: drawSphere(); break;
	case MODEL_TORUS: drawTorus(); break;
	case MODEL_TEAPOT: drawTeapot(); break;
	case MODEL_PLANE: drawPlane(); break;
	};
	glPopMatrix();
}

void drawObject::animate(const drawWorld &w,double dt) 
{
	orientation.nudge(dt*angVel.x,dt*angVel.y); // HACK for rotation
	
	velocity.z-=dt*9.8; /* Gravitational accelleration pulls velocity downward */
	center+=dt*velocity;
	if (center.z<w.groundZ+radius && velocity.z<0) {
		center.z=w.groundZ+radius;
		velocity.z*=-0.93;
		velocity*=0.98; /* lose energy at each hit */
		angVel.y=-angVel.y; /* Spin along a different axis */
		angVel*=0.4;
	}
}


/* ShadowMethod interface destructor */
shadowMethod::~shadowMethod() {}

/****************** No Shadows *******************/
class noShadows : public shadowMethod {
public:
	virtual void draw(const drawWorld &w) {
		w.drawScene(); /* Draw world without shadows */
	}
};

/***************** Decal Shadows *****************/
/* Draws a fixed texture "decal" shadow underneath each object. */
class decalShadows : public shadowMethod {
public:
	oglTexture *decal; /* shadow decal texture */
	
	decalShadows() {
		/* Make a new shadow decal texture: a central dark spot
		  with fuzzy edges.  This is a greyscale texture.
		*/
		const int w=32, h=32;
		unsigned char decalArr[w*h];
		for (int y=0;y<h;y++)
		for (int x=0;x<w;x++) {
			double dx=x/(double)w-0.5, dy=y/(double)h-0.5;
			double r=sqrt(dx*dx+dy*dy);
			double out=1.0;
			double r_min=0.2, r_max=0.4;
			if (r<r_min) out=0;
			else if (r>r_max) out=1.0;
			else out=(r-r_min)/(r_max-r_min);
			decalArr[y*w+x]=(unsigned char)(255*(0.8*out+0.2));
		}
		decal=new oglTexture(decalArr,w,h,
			oglTexture_mipmap, GL_LUMINANCE,GL_UNSIGNED_BYTE,
			GL_LUMINANCE8);
	}

	virtual void draw(const drawWorld &w);
	virtual void shadowObjects(const drawWorld &w,const mat4 &toGround);
	virtual void shadowObject(const drawWorld &w,const drawObject &o,const mat4 &toGround);
};

void decalShadows::draw(const drawWorld &w) {
	set_shadow_detail_names("Show shadow centerline");
	w.drawScene();
	
	shadowObjects(w,w.lightFlatten);
	shadowObjects(w,w.lightFlatten2);
}
void decalShadows::shadowObjects(const drawWorld &w,const mat4 &toGround)
{
	/* Turn off fancy pixel shader--use fixed-function OpenGL */
        glUseProgramObjectARB(0);
	for (unsigned int i=0;i<w.objects.size();i++) 
		shadowObject(w,w.objects[i],toGround);
}
// Draw a decal shadow underneath this object
void decalShadows::shadowObject(const drawWorld &w,const drawObject &o,const mat4 &toGround)
{
	/* Project the object center down onto the ground */
	vec4 worldCenter=w.modelMat*vec4(o.center,1.0);
	vec4 shadowCenter=toGround*worldCenter;
	
	if (shadow_detail1) { /* draw line down shadow centers */
		glColor4f(1,1,0,0.5);
		glBegin(GL_LINES);
		glVertex4fv(worldCenter); glVertex4fv(shadowCenter);
		glEnd();
	}
	
	/* Compute the approximate size of the shadow quad */
	double s=1.7*o.radius;
	
	/* Draw one quad that darkens the ground beneath it */
	glColor4f(1,1,1,1);
	glBlendFunc(GL_DST_COLOR, GL_ZERO); /* out=src*dest */
	glDepthMask(GL_FALSE); /* don't mess up depth buffer */
	osl::Vector3d v=osl::Vector3d((float *)shadowCenter);
	oglTextureQuad(*decal,
		  v+s*osl::Vector3d(-1,+1,0),v+s*osl::Vector3d(+1,+1,0),
		  v+s*osl::Vector3d(-1,-1,0),v+s*osl::Vector3d(+1,-1,0));
	glDepthMask(GL_TRUE);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); /* back to normal alpha blending */
}


/****************** Projected Shadows *******************/
/** Squishes all geometry onto the ground plane */
class projectedShadows : public shadowMethod {
public:
	virtual void draw(const drawWorld &w);
	virtual void squishAndDraw(const drawWorld &w,const mat4 &toGround);
};

// Draw world's geometry, projected to the ground plane.
void projectedShadows::draw(const drawWorld &w) {
	set_shadow_detail_names("","","Enable depth writes");

/* Draw world without shadows */
	w.drawScene();

/* Draw world squished to ground along the lightFlatten matrix*/
	squishAndDraw(w,w.lightFlatten);
	squishAndDraw(w,w.lightFlatten2);
}

void projectedShadows::squishAndDraw(const drawWorld &w,const mat4 &toGround) {
/* Squish object geometry onto ground plane */
	mat4 m=w.persMat*w.viewMat*toGround;
	static oglProgramObject *p=oglProgramFromStrings(
	"/* Vertex program */\n"
	"uniform mat4 m;\n"
	"void main(void) { gl_Position=m*(gl_ModelViewMatrix*gl_Vertex); }\n"
	,
	"/* Fragment program */\n"
	"void main(void) { gl_FragColor=vec4(0.0,0.0,0.0,0.5);}\n"
	);
	p->begin();
	p->set("m",m);
	if (!shadow_detail3) glDepthMask(GL_FALSE); /* don't mess up depth buffer */
	w.drawObjects();
	if (!shadow_detail3) glDepthMask(GL_TRUE);
	p->end();
}

/****************** Projected Shadows with Stencil *******************/
class projectedStencilShadows : public projectedShadows {
public:
	virtual void draw(const drawWorld &w);
	virtual void stencilSquish(const drawWorld &w,const mat4 &toGround);
};

void projectedStencilShadows::draw(const drawWorld &w) {
	set_shadow_detail_names("Show double-shadow locations","Show stencil buffer","Enable depth writes");
	
// Draw world without shadows
	w.drawScene();
	
// Draw shadows
	stencilSquish(w,w.lightFlatten);
	stencilSquish(w,w.lightFlatten2);
}

/* Draw a whole-screen black quad with this alpha value */
void screenQuad(double alphaIntensity) {
	glColor4f(0,0,0,alphaIntensity);
	glDisable(GL_DEPTH_TEST);
	oglDrawScreenQuad(); /* fill whole screen with black (where stencil test passes!) */
	glEnable(GL_DEPTH_TEST);
}

void projectedStencilShadows::stencilSquish(const drawWorld &w,const mat4 &toGround)
{
	glClear(GL_STENCIL_BUFFER_BIT);
	glEnable(GL_STENCIL_TEST); /* needed for stencil buffer *writes* to work! */
	
// Set the stencil for all the projected geometry
	glStencilFunc(GL_ALWAYS,1,0xff);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
	glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE); /* don't mess with color */
	
// Draw squished geometry to stencil buffer
	squishAndDraw(w,toGround);
	
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE); /* allow changes to color */
	if (shadow_detail2) return; /* give up--show stencil buffer so far */
// Darken the shadows, by drawing where the 0 is less than the stencil
	glStencilFunc(GL_NOTEQUAL,shadow_detail1?2:0,0xff); 
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	screenQuad(0.5);
	glDisable(GL_STENCIL_TEST);
}

/****************** Shadow Volumes ('V') ******************/

class shadowVolumes : public shadowMethod {
public:
	virtual void draw(const drawWorld &w);
	
	/* Draw shadows of light from this direction */
	virtual void drawShadows(const drawWorld &w,const vec3 &lightDir);
	
	/* Draw shadow silhouttes for all the geometry in the world, from this light direction */
	virtual void drawSilhouettes(const drawWorld &w,const vec3 &lightDir,GLenum drawSide,bool endCaps);
};

void shadowVolumes::draw(const drawWorld &w)
{
	set_shadow_detail_names("Show shadow volumes","Use Carmack depth-flip","Skip back faces");
// Draw geometry normally
	w.drawScene();
	
// Draw shadows
	drawShadows(w,w.lightDir);
	drawShadows(w,w.lightDir2);
}

/* Draw shadows of light from this direction */
void shadowVolumes::drawShadows(const drawWorld &w,const vec3 &lightDir)
{
// Set up stencil buffer with location of shadows
	glClear(GL_STENCIL_BUFFER_BIT);
	glUseProgramObjectARB(0); /* Turn off fancy pixel shader--all fixed-function OpenGL */
	glDepthMask(GL_FALSE); /* don't mess up z buffer */
	glEnable(GL_STENCIL_TEST); /* needed for stencil buffer *writes* to work! */
	glStencilFunc(GL_ALWAYS,1,0xff);
	
	/* Render shadow silhouettes into stencil buffer */
	if (shadow_detail1) {
		/* Draw silhouettes onscreen, in yellow */
		glColor4f(1,1,0,0.2);
	} else 
	{ /* Regular shadow volumes */
		glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE); /* don't mess with color */
	}
	
	if (!shadow_detail2) 
	{ /* Normal shadow volumes: */
		/* Increment all the front faces */
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR); 
		drawSilhouettes(w,lightDir,GL_FRONT,true);
		/* Decrement all the back faces */
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		drawSilhouettes(w,lightDir,GL_BACK,true);
	} 
	else
	{ /* Use Carmack reversing trick */
		glDepthFunc(GL_GREATER); /* flip the sense of the depth compare */
		/* Increment all the back faces that are hidden */
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR); 
		drawSilhouettes(w,lightDir,GL_BACK,true);
		/* Decrement all the front faces that are hidden */
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		drawSilhouettes(w,lightDir,GL_FRONT,true);
		glDepthFunc(GL_LESS);
	} 
	
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);

// Darken the shadows, by drawing where the stencil is >0 (that is, where we're in shadow)
//  FIXME: chould instead first draw ambient, then add diffuse where stencil is *0*.
	if (!shadow_detail1) {
		glStencilFunc(GL_NOTEQUAL,0,0xff); /* draw where stencil is not zero */
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP); 
		screenQuad(0.5);
	}
	
	glDisable(GL_STENCIL_TEST);
	glDepthMask(GL_TRUE);
}

silhoutteConsumer::~silhoutteConsumer() {}

// Receives pieces of a shadow silhouette
class silhoutteAccum : public silhoutteConsumer {
public:
	// World-coordinate points along shadow silhouttes
	std::vector<vec3> sil;
	// Indices into sil vector giving first index of each silhoutte curve
	std::vector<int> starts;
	
	// Here's the first point on a shadow silhoutte.
	//   world is the vertex location in world coords
	virtual void silhoutte_first(const vec3 &world) {starts.push_back(sil.size()); sil.push_back(world);}
	// Here's the next point in a shadow silhoutte.
	virtual void silhoutte_next(const vec3 &world) {sil.push_back(world);}
};

void shadowVolumes::drawSilhouettes(const drawWorld &w,const vec3 &lightDir,GLenum drawSide,bool endCaps) 
{
	/* Expensive hack: draws all faces, but culls out those facing wrong way! */
	glEnable(GL_CULL_FACE);
	if (shadow_detail3) { /* don't draw back sides */
		if (drawSide==GL_BACK) return;
		if (drawSide==GL_FRONT_AND_BACK) drawSide=GL_FRONT;
	}
	switch (drawSide) {
	case GL_FRONT: glCullFace(GL_BACK); break; // to draw front, cull back
	case GL_BACK: glCullFace(GL_FRONT); break; // to draw back, cull front
	case GL_FRONT_AND_BACK: glDisable(GL_CULL_FACE); break;
	}
	for (unsigned int s, i=0;i<w.objects.size();i++) {
		silhoutteAccum ac;
		w.objects[i].getSilhouette(w,lightDir,ac);
		/* Extrude silhouttes listed in ac */
		glBegin(GL_TRIANGLE_STRIP);
		for (s=0;s<ac.sil.size();s++) {
			glVertex3fv(ac.sil[s]-20.0*lightDir);
			glVertex3fv(ac.sil[s]);
		}
		glEnd();
		if (endCaps) { /* Draw caps on start and end of silhouttes */
			glBegin(GL_TRIANGLE_FAN); /* cap on end */
			for (s=0;s<ac.sil.size();s++) glVertex3fv(ac.sil[s]-20.0*lightDir);
			glEnd();
			
			glBegin(GL_TRIANGLES); /* cap on start-- gotta flip orientation... */
			for (s=0;s<ac.sil.size()-1;s++) {
				glVertex3fv(ac.sil[0]);
				glVertex3fv(ac.sil[s+1]);
				glVertex3fv(ac.sil[s]);
			}
			glEnd();
		}
	}
	glDisable(GL_CULL_FACE);
}

// Extract this object's silhoutte points--the places where the surface is perpendicular to perpDir.
//   FIXME: eventually this should be made to work with non-sphere objects.
void drawObject::getSilhouette(const drawWorld &w,const vec3 &lightDir,silhoutteConsumer &dest) const
{
	vec3 worldCenter=vec3(w.modelMat*vec4(center,1.0));
	vec3 up=lightDir.cross(vec3(0,1,0)).dir(); /* Pick a random direction perpendicular to the light */
	vec3 rt=lightDir.cross(up).dir();
	double del=2*M_PI/16; /* divide up circumference of sphere into points */
	vec3 first=vec3(0.0);
	for (double ang=0;ang<2*M_PI;ang+=del) {
		vec3 world=worldCenter+radius*(up*sin(ang)+rt*cos(ang));
		if (ang==0) {first=world;dest.silhoutte_first(world);}
		else dest.silhoutte_next(world);
	}
	dest.silhoutte_next(first); /* close the loop around the sphere's surface */
}

/********************* Shadow Maps **********************/

/** Prepare a matrix with origin "origin", 1,1,1 position "unit",
and axes axX, axY, and axZ.  That is, the matrix will map:
	vec4(0,0,0,1) == m*origin
	vec4(1,1,1,1) == m*unit
	vec4(1,0,0,0) == normalize(m*axX);
	vec4(0,1,0,0) == normalize(m*axY);
	vec4(0,0,1,0) == normalize(m*axZ);
*/
mat4 makeMatrix(vec3 origin,vec3 unit, vec3 axX,vec3 axY,vec3 axZ) {
	mat4 m=mat4(1.0);
	/* Find rows of matrix.  
	   Cross-products are to match the zeros in axX,Y,Z products above;
	   scaling is so m*del==vec4(1.0).
	*/
	vec3 del=unit-origin;
	vec3 rowX=axY.cross(axZ).dir(); rowX*=1.0/dot(rowX,del); 
	vec3 rowY=axZ.cross(axX).dir(); rowY*=1.0/dot(rowY,del); 
	vec3 rowZ=axX.cross(axY).dir(); rowZ*=1.0/dot(rowZ,del); 
	for (int col=0;col<3;col++) {
		m[col].x=rowX[col];
		m[col].y=rowY[col];
		m[col].z=rowZ[col];
	}
	/* Choose constants so m*origin is 0.0 */
	m[3]=vec4(-dot(rowX,origin),-dot(rowY,origin),-dot(rowZ,origin),1.0);
	return m;
}

class shadowMapParameters {
public:
	int shadow_wid, shadow_ht;
	vec3 span;
	vec3 shadowZspan;
	vec3 sh0;/* 0,0,0 in shadow texture coords */
	vec3 sh1;/* 1,1,1 in shadow texture coords */
	vec3 shX,shY,shZ; /* axes of shadow texture coords */
	
	/* This matrix maps world coordinates into shadow map texture coordinates */
	mat4 texFmWorld;
	
	shadowMapParameters(const vec3 &lightDir,const vec3 &span,double groundZ);
	void renderCoords(const drawWorld &w);
};
shadowMapParameters::shadowMapParameters(const vec3 &lightDir,const vec3 &span_,double groundZ) {
	span=span_;
	switch(shadowmap_res) {
	case 0: shadow_wid=shadow_ht=512; break;
	case 1: shadow_wid=shadow_ht=128; break;
	case 2: shadow_wid=shadow_ht=32; break;
	}
	
	vec3 lightDirUnitZ=lightDir/lightDir.z; /* scaled so lightDirUnitZ.z==1.0 */
	shadowZspan=lightDirUnitZ*span.z;
	sh0=vec3(-span.x,-span.y,groundZ)+shadowZspan; 
	sh1=vec3(+span.x,+span.y,groundZ); 
	shX=vec3(1,0,0); /* +X axis */
	shY=vec3(0,1,0); /* +Y axis */
	shZ=lightDir; /* +Z axis */
	
	texFmWorld=makeMatrix(sh0,sh1, shX,shY,shZ);
}
/* For debugging: show our coordinates */
void shadowMapParameters::renderCoords(const drawWorld &w) {
	static oglProgramObject *p=oglProgramFromStrings(
		"/* Vertex program */\n"
		"uniform mat4 clipFmWorld, texFmWorld;\n"
		"varying vec4 shadowCoords;\n"
		"void main(void) { \n"
		"	vec4 world=gl_ModelViewMatrix*gl_Vertex; /* world coords */ \n"
		"	gl_Position=clipFmWorld*world; \n"
		"	shadowCoords=texFmWorld*world; \n"
		"}\n"
	,
		"/* Fragment program */\n"
		"varying vec4 shadowCoords;\n"
		"void main(void) { gl_FragColor=fract(shadowCoords); gl_FragColor.a=1.0;}\n"
	);
	p->begin();
	p->set("clipFmWorld",w.persMat*w.viewMat); /* normal camera viewpoint */
	p->set("texFmWorld",texFmWorld);
	w.drawScene();
	p->end();
}

class shadowMaps : public shadowMethod {
public:
	/* Most recent shadow map (saving this texture is an optimization) */
	oglTexture *shadowMap, *shadowMap2;
	shadowMaps();
	
	virtual void draw(const drawWorld &w);
	virtual void makeShadowMap(const drawWorld &w,const shadowMapParameters &sp,oglTexture *&dest);
	virtual void drawWithShadows(const drawWorld &w,const shadowMapParameters &sp,const shadowMapParameters &sp2);
};

shadowMaps::shadowMaps() {
	shadowMap=shadowMap2=0;
}

void shadowMaps::draw(const drawWorld &w) {
	set_shadow_detail_names("Show shadow map","Show light-source view","Show shadow-map coordinates");
	
	vec3 span=vec3(1.5,1.5,3.0); /* per-axis size of shadow map */
	shadowMapParameters sp=shadowMapParameters(w.lightDir,span,w.groundZ);
	shadowMapParameters sp2=shadowMapParameters(w.lightDir2,span,w.groundZ);
	if (shadow_detail3) {sp.renderCoords(w); return;}
	
	makeShadowMap(w,sp2,shadowMap2);
	makeShadowMap(w,sp,shadowMap);
	
	if (shadow_detail2) return; /* Show off raw light-source view of world */
	
	drawWithShadows(w,sp,sp2);
}

void shadowMaps::makeShadowMap(const drawWorld &w,const shadowMapParameters &sp,oglTexture *&dest)
{
/* Parameterize the shadow map--figure out how big it is, etc. */
	int shadow_x=0, shadow_y=win_h-sp.shadow_ht; /* avoid bottom of screen, where GUI crap sits! */
	
/* Rasterize world's depths, from light source point of view, into depth buffer */
	glPushAttrib(GL_ALL_ATTRIB_BITS); /* save everything before making depth image */
	glViewport(shadow_x,shadow_y,sp.shadow_wid,sp.shadow_ht);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	static oglProgramObject *shP=oglProgramFromStrings(
		"/* Vertex program */\n"
		"uniform mat4 clipFmWorld;\n"
		"void main(void) { \n"
		"	vec4 world=gl_ModelViewMatrix*gl_Vertex; /* world coords */ \n"
		"	gl_Position=clipFmWorld*world; \n"
		"}\n"
	,
		"/* Fragment program-- all we need is depth; might as well draw purple objects! */\n"
		"void main(void) { gl_FragColor=vec4(1.0,0.0,1.0,1.0);}\n"
	);
	shP->begin();
	
	/* This matrix maps texture map coords on [0,1] into clip coordinates in [-1,+1] 
		clip = tex*2-1
	*/
	mat4 clipFmTex=mat4(2.0);
	clipFmTex[3]=vec4(-1,-1,-1,1);
	
	shP->set("clipFmWorld",clipFmTex*sp.texFmWorld);
	w.drawObjects();
	shP->end();

/* Read back computed depth image */
	w.rb->readDepth(shadow_x,shadow_y,sp.shadow_wid,sp.shadow_ht);
	static int last_size=0;
	if (dest==0 || last_size!=sp.shadow_wid) {
		last_size=sp.shadow_wid;
		delete dest;
		dest=w.rb->makeTex();
	}
	w.rb->readTex(dest);
	glPopAttrib(); /* to restore viewport, mostly... */
}

void shadowMaps::drawWithShadows(const drawWorld &w,const shadowMapParameters &sp,const shadowMapParameters &sp2) {
	unsigned short *depth=0; /* <- shadowmap depths, copied to CPU */
	if (shadow_detail1) 
	{ /* Turn shadow map depths into geometry */
		depth=(unsigned short *)w.rb->getData();
	}
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
/* Render world from camera POV, using computed depths */
	static oglProgramObject *camP=oglProgramFromStrings(
		"/* Vertex program */\n"
		"uniform mat4 clipFmWorld, texFmWorld, texFmWorld2;\n"
		"varying vec4 shadowCoords, shadowCoords2, color;\n"
		"varying vec3 normal;\n"
		"void main(void) { \n"
		"	vec4 world=gl_ModelViewMatrix*gl_Vertex; /* world coords */ \n"
		"	gl_Position=clipFmWorld*world; \n"
		"	shadowCoords = texFmWorld*world;\n"
		"	shadowCoords2 = texFmWorld2*world;\n"
		"	color=gl_Color;\n"
		"	normal=normalize(gl_NormalMatrix*gl_Normal);\n"
		"}\n"
	,
		"/* Fragment program-- use shadow map to determine shadowed points */\n"
		"varying vec4 shadowCoords, shadowCoords2, color;\n"
		"varying vec3 normal;\n"
		"uniform vec4 lightDir, lightDir2;\n"
	#if 0 /* Unfiltered shadows */
		"uniform sampler2D shadowMap;\n"
		"void main(void) { \n"
		"	float shadow=step(shadowCoords.z,texture2D(shadowMap,vec2(shadowCoords)).x);\n"
	#else /* Reeves filtering (doesn't seem to change anything on my laptop...) */
		"uniform sampler2DShadow shadowMap;\n"
		"uniform sampler2DShadow shadowMap2;\n"
		"void main(void) { \n"
		"	float shadow=shadow2D(shadowMap,vec3(shadowCoords)).x;\n"
		"	float shadow2=shadow2D(shadowMap2,vec3(shadowCoords2)).x;\n"
	#endif
		"// usual per-pixel lighting\n"
		"	vec3 N=normalize(normal);\n"
		"	float dotNL=clamp(dot(N,vec3(lightDir)),0.0,1.0);\n"
		"	float dotNL2=clamp(dot(N,vec3(lightDir2)),0.0,1.0);\n"
		"	gl_FragColor=color*(0.1+0.4*shadow*dotNL+0.4*shadow2*dotNL2);\n"
		"	gl_FragColor.a=1.0;\n"
		"}\n"
	);
	camP->begin();
	camP->set("clipFmWorld",w.persMat*w.viewMat); /* normal camera viewpoint */
	camP->set("lightDir",w.lightDir);
	camP->set("lightDir2",w.lightDir2);
	/* Bias geometry's location upward *slightly* to avoid self-shadowing */
	mat4 texFmWorld=sp.texFmWorld; texFmWorld[3].z-=0.002;
	camP->set("texFmWorld",texFmWorld); /* texture coordinates from world coordinates */
	
	mat4 texFmWorld2=sp2.texFmWorld; texFmWorld2[3].z-=0.002;
	camP->set("texFmWorld2",texFmWorld2); /* texture coordinates from world coordinates */
	
	glActiveTextureARB(GL_TEXTURE1_ARB);
	camP->set("shadowMap2",1); /* texture 1 */
	shadowMap2->bind();
	
	glActiveTextureARB(GL_TEXTURE0_ARB);
	camP->set("shadowMap",0); /* texture 0 */
	shadowMap->bind();
	
	/* enable blending on shadowMap depth texture */
	oglTextureType(oglTexture_linear); 
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_COMPARE_MODE,GL_COMPARE_R_TO_TEXTURE_ARB);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_COMPARE_FUNC,GL_LEQUAL);
	
	w.drawScene();
	camP->end();
	
	if (depth) 
	{ /* debug 3: Turn shadow map depths into geometry */
		float delX=2*sp.span.x/sp.shadow_wid; /* X size of a shadowmap pixel */
		float delY=2*sp.span.y/sp.shadow_ht; /* Y size of a shadowmap pixel */
		vec3 zFmDepth=-sp.shadowZspan/0xffffu; /* Z size of a shadowmap depth unit */
		for (int y=sp.shadow_ht-2;y>=0;y--) {
			glColor4f(1,(y%2)?1:0.7,0,0.5);
			glBegin(GL_TRIANGLE_STRIP);
			vec3 loY=sp.sh0+vec3(delX*0.5,delY*(y+0.5),0.002); /* world coordinates of start of row of shadowmap pixels */
			vec3 hiY=sp.sh0+vec3(delX*0.5,delY*(y+1.5),0.002);
			for (int x=0;x<sp.shadow_wid-1;x++)
			{
				glVertex3fv(depth[(y+0)*sp.shadow_wid+x]*zFmDepth+loY);
				glVertex3fv(depth[(y+1)*sp.shadow_wid+x]*zFmDepth+hiY);
				loY.x+=delX; hiY.x+=delX;
			}
			glEnd();
		}
	}
}

/************************** Penumbra Limit Maps ***************************/
/* FIXME: actually make this class function */
class penumbraLimitMaps : public shadowMaps {
};

/****************************** End of Lighting ********************/

/******** Program Initialization *********/
drawWorld *mainObject=0; /**< Big object we're trying to draw */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	static std::vector<shadowMethod *> shadows;
	if (shadows.size()==0) {
		shadows.push_back(new noShadows);
		shadows.push_back(new decalShadows);
		shadows.push_back(new projectedShadows);
		shadows.push_back(new projectedStencilShadows);
		shadows.push_back(new shadowVolumes);
		shadows.push_back(new shadowMaps);
		shadows.push_back(new penumbraLimitMaps);
	}

	Check("Entering display routine");
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT); /* actually clear color and depth */
	
	// Set up a little error-prone OpenGL state:
	glEnable(GL_BLEND); /* alpha blending */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw(*shadows[shadow_mode]);
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w, win_h=h;
	Check("Before glViewport");  
	int tx, ty, tw, th;
	GLUI_Master.get_viewport_area( &tx, &ty, &tw, &th );
	glViewport( tx, ty, view_w=tw, view_h=th );
	Check("After glViewport");
}

void make_list(GLUI *glui,const char *title,int *mode_ptr,const char **names)
{
	GLUI_Listbox *list = new GLUI_Listbox( glui, title, mode_ptr);
	for (int i=0;names[i];i++) 
		list->add_item(i,names[i]);
	list->do_selection(*mode_ptr);
}

int main(int argc, char** argv)
{
	glutInitWindowSize(900,750);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");
	
	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Make an object to draw
	mainObject=new drawWorld;
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_BOTTOM);
	glui->set_main_gfx_window( main_window );
	

	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	//glutReshapeFunc (main_reshape);
	GLUI_Master.set_glutReshapeFunc( main_reshape );  
	
	make_list(glui,"Shadows:",&shadow_mode,shadow_mode_names);
	check_detail1=new GLUI_Checkbox(glui,"Detail 1:",&shadow_detail1);
	check_detail2=new GLUI_Checkbox(glui,"Detail 2:",&shadow_detail2);
	check_detail3=new GLUI_Checkbox(glui,"Detail 3:",&shadow_detail3);
	make_list(glui,"Resolution:",&shadowmap_res,shadowmap_res_names);

	//new GLUI_Column(glui);
	//make_list(glui,"Model:",&model_mode,model_mode_names);
	
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Model:",&gui_modelMat[0][0]);
	make_list(glui,"Type:",&model_mode,model_mode_names);
	make_list(glui,"Animate:",&model_anim,model_anim_names);
	new GLUI_Column(glui,false);
	
	new GLUI_Rotation(glui,"View:",&gui_viewMat[0][0]);
	GLUI_Spinner *spin_d = new GLUI_Spinner( glui, "Distance:", &cameraDist);
	spin_d->set_float_limits(0.1, 10.0);
	GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "Zoom:", &cameraZoom);
	spin_zoom->set_float_limits(0.1, 10.0);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"Light1:",&light1Mat[0][0]);
	new GLUI_Rotation(glui,"Light2:",&light2Mat[0][0]);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
